from iktomi.db.sqla.images import * # pragma: no cover
import warnings # pragma: no cover
warnings.warn('iktomi.unstable.db.sqla.images is deprecated. Use ' # pragma: no cover
              'iktomi.db.sqla.images instead',
              category=DeprecationWarning,
              stacklevel=2)
