from iktomi.forms.files import * # pragma: no cover
import warnings # pragma: no cover
warnings.warn('iktomi.unstable.forms.files is deprecated. Use ' # pragma: no cover
              'iktomi.forms.files instead',
              category=DeprecationWarning)
