from iktomi.db.sqla.files import * # pragma: no cover
import warnings # pragma: no cover
warnings.warn('iktomi.unstable.db.sqla.files is deprecated. Use ' # pragma: no cover
              'iktomi.db.sqla.files instead',
              category=DeprecationWarning)
