<table class="fieldlist" id="{{ field.input_name }}">
    <tbody>
    {% for index in field.form.raw_data.getall(field.indices_input_name) %}
        <tr class="fieldlist-item">
            <td class="fieldlist-cell">
                <input type="hidden" name="{{ field.indices_input_name }}" value="{{ index }}" />
                {% set f = field.field(name=index) %}
                {% if f.error %}
                <div>
                  <span class="error">{{ f.error }}</span>
                </div>
                {% endif %}
                {{ f.widget.render() }}
            </td>
            {% if field.order %}
            <td class="fieldlist-order fieldlist-btns">&nbsp;</td>
            {% endif %}
            <td class="fieldlist-delete fieldlist-btns">&nbsp;</td>
        </tr>
    {% endfor %}
    </tbody>
</table>
