<input id="{{ widget.id }}" type="text" name="{{ widget.input_name|escape }}" value="{{ value|forceescape }}"
       {%- if readonly %} readonly="readonly"{% endif %}
       {%- if widget.classname %} class="{{ widget.classname }}"{% endif %} />
