<div {%- if widget.classname %} class="{{ widget.classname }}"{% endif %}>
  {% for option in options %}
    <div>
       <input type="{{ 'checkbox' if widget.multiple else 'radio' }}" value="{{ option.value|forceescape }}" {#
              #}name="{{ widget.input_name|escape }}" id="{{ widget.id }}-{{ loop.index }}"
              {%- if option.selected %} checked="checked"{% endif %} />
       <label for="{{ widget.id }}-{{ loop.index }}"
              {%- if option.selected %} class="selected"{% endif %}>
            {{- option.title|escape -}}
       </label>
    </div>
  {%- endfor %}
</div>
