{% for field in form.fields %}
{% if field.readable %}
<p class="form-row" {% if field.widget.render_type=="hidden" %}style="display: none"{% endif %}>
  {% if field.error %}
  <span class="error">{{ field.error }}</span>
  {% endif %}
  {% if field.widget.render_type == 'checkbox' %}
    {{ field.widget.render()|safe }}
    {% if field.label %}
    <label for="{{ field.id }}">{{ field.label }}</label>
    {% endif %}
  {% elif field.widget.render_type == 'hidden' %}
    {{ field.widget.render()|safe }}
  {% else %}
    {% if field.label %}
    <label for="{{ field.id }}">{{ field.label }}</label>
    {% endif %}
    {{ field.widget.render()|safe }}
  {% endif %}

  {%- if field.hint and not field.widget.renders_hint  %}
    <span class="hint">{{ field.hint }}</span>
  {%- endif -%}
</p>
{% endif %}
{% endfor %}
